from jovian._version import __version__
from time import sleep
from jovian.utils.anaconda import upload_conda_env, CondaError
from jovian.utils.pip import upload_pip_env
from jovian.utils.api import create_gist_simple, upload_file, get_gist
from jovian.utils.logger import log
from jovian.utils.constants import WEBAPP_URL, FILENAME_MSG, RC_FILENAME
from jovian.utils.jupyter import set_notebook_name, in_notebook, save_notebook, get_notebook_name
from jovian.utils.rcfile import get_notebook_slug, set_notebook_slug, make_rcdata

set_notebook_name()

_current_slug = None


def commit(secret=False, nb_filename=None, capture_env=True, env_type='conda', notebook_id=None):
    """Save the notebook, capture environment and upload to the cloud for sharing.

    In most cases, commit works well with the default arguments. It attempts to 
    1. Save the Jupyter notebook
    2. Upload the notebook to https://jvn.io
    3. Capture the python environment (using Anaconda or pip)
    4. Upload the python environment to cloud

    Capturing the python environment ensures that the notebook can be reproduced and 
    executed easily using the `jovian` command line tool. For more details, see 
    https://jvn.io/getting-started . 

    Issues and bugs can be reported here: https://github.com/swiftace-ai/jovian-py

    Arguments:

        secret (bool, optional): Create a secret notebook on https://jvn.io , which is only 
            accessible via the link, and is not visible on the owner's public profile. By default,
            commited notebooks are public and visible on the owner's profile.

        nb_filename (string, optional): The filename of the jupyter notebook (including 
            the .ipynb extension). This is detected automatically in most cases, but in
            certain environments like Jupyter Lab, the detection may fail and the filename
            needs to be provided using this argument.

        capture_env (bool, optional): If `True`, the Python environment (python version,
            libraries etc.) are captured and uploaded along with the notebook.

        env_type (string, optional): The type of environment to be captured. Allowed options are
            'conda' and 'pip'.

        notebook_id (string, optional): If you wish to update an existing notebook owned by you,
            you can use this argument to provide the base64 ID (present in the URL) of an notebook 
            hosted on https://jvn.io . In most cases, this argument is not required, and the library
            can automatically infer whether you are looking to update an existing notebook or create
            a new one.

    """
    global _current_slug

    # Check if we're in a Jupyter environment
    if not in_notebook():
        log('Failed to detect Juptyer notebook. Skipping..', error=True)
        return

    # Save the notebook (uses Javascript, doesn't work everywhere)
    log('Saving notebook..')
    save_notebook()
    sleep(1)

    # Get the filename of the notebook (if not provided)
    if nb_filename is None:
        nb_filename = get_notebook_name()

    # Exit with help message if filename wasn't detected (or provided)
    if nb_filename is None:
        log(FILENAME_MSG)
        return

    # Check whether to create a new gist, or update an old one
    if notebook_id is None:
        # First preference to the in-memory slug variable
        if _current_slug is not None:
            notebook_id = _current_slug
        else:
            notebook_id = get_notebook_slug(nb_filename)

    # Check if the current user can push to this slug
    if notebook_id is not None:
        gist_meta = get_gist(notebook_id)
        if not gist_meta['isOwner']:
            notebook_id = None

    # Log whether this is an update or creation
    if notebook_id is None:
        log('Creating a new notebook on https://jvn.io')
    else:
        log('Updating notebook "' + notebook_id + '" on https://jvn.io')

    # Upload the notebook & create/update the gist
    res = create_gist_simple(nb_filename, notebook_id, secret)
    if res is None:
        return

    # Extract slug and owner from created gist
    slug, owner = res['slug'], res['owner']

    # Set/update the slug information
    _current_slug = slug
    set_notebook_slug(nb_filename, slug)

    # Save & upload environment
    if capture_env:
        log('Capturing environment..')

        if env_type == 'conda':
            # Capture conda environment
            try:
                upload_conda_env(slug)
            except CondaError as e:
                log(str(e), error=True)

        elif env_type == 'pip':
            # Capture pip environment
            try:
                upload_pip_env(slug)
            except Exception as e:
                log(str(e), error=True)

    # Print commit URL
    log('Committed successfully! ' + WEBAPP_URL +
        "/" + owner['username'] + "/" + slug)



"""Anaconda related utilities"""
import os
import logging
from jovian.utils.api import upload_file
from jovian.utils.misc import get_platform
from jovian.utils.constants import PLATFORMS


class CondaError(Exception):
    """Error class for Anaconda-related exceptions"""
    pass


CONDA_NOT_FOUND = 'Anaconda binary not found. Please make sure the "conda" command is in your system PATH or the environment variable $CONDA_EXE points to the anaconda binary'


def get_conda_bin():
    """Get the path to the Anaconda binary"""
    conda_bin = 'conda'
    # Try executing the `conda` command
    if os.popen(conda_bin).read().strip() == '':
        # If it fails, look for $CONDA_EXE
        conda_exe = os.popen('echo $CONDA_EXE').read().strip()
        # Check if it returns a valid path
        if conda_exe != '' and conda_exe != '$CONDA_EXE':
            # Update binary and execute again
            conda_bin = conda_exe
            if os.popen(conda_bin).read().strip() == '':
                raise CondaError(CONDA_NOT_FOUND)
    logging.info('Anaconda binary: ' + conda_bin)
    return conda_bin


def get_conda_env_name():
    """Get the name of the active conda environment"""
    env_name = os.popen('echo $CONDA_DEFAULT_ENV').read().strip()
    if env_name == '' or env_name == '$CONDA_DEFAULT_ENV':
        env_name = 'base'
    logging.info('Anaconda environment: ' + env_name)
    return env_name


def read_conda_env(name=None):
    """Read the anaconda environment into a string"""
    if name is None:
        name = get_conda_env_name()
    command = get_conda_bin() + ' env export -n ' + \
        get_conda_env_name() + " --no-builds"
    env_str = os.popen(command).read()
    if env_str == '':
        error = 'Failed to read Anaconda environment using command: "' + command + '"'
        raise CondaError(error)
    return env_str


def upload_conda_env(gist_slug):
    """Read and save the current Anaconda environment to server"""
    # Export environment to YML string
    env_str = read_conda_env(get_conda_env_name())

    # Upload environment.yml
    upload_file(gist_slug, ('environment.yml', env_str))

    # Check and include existing os-specific files
    platform = get_platform()
    for p in PLATFORMS:
        pfname = 'environment-' + p + '.yml'
        if p == platform:
            # Use the new environment for current platform
            upload_file(gist_slug, (pfname, env_str))
        elif os.path.exists(pfname):
            # Reuse old environments for other platforms
            upload_file(gist_slug, pfname)



import argparse
import webbrowser
from jovian.utils.clone import clone, pull
from jovian.utils.install import install, activate
from jovian._version import __version__


def exec_clone(slug):
    clone(slug)


def exec_init():
    from jovian.utils.api import get_api_key
    from jovian.utils.credentials import get_guest_key
    # webbrowser.open('https://jvn.io/')
    get_guest_key()
    get_api_key()
    print('Initialization finished')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('command')
    parser.add_argument('gist', nargs='?')

    args = parser.parse_args()
    command = args.command
    if command == 'init':
        exec_init()
    elif command == 'clone':
        if not args.gist:
            print('Please provide the Gist ID to clone')
            return
        exec_clone(args.gist)
    elif command == 'pull':
        pull(args.gist)
    elif command == 'version':
        print('Jovian library version: ' + __version__)
    elif command == 'install':
        install()
    elif command == 'activate':
        activate()


if __name__ == '__main__':
    main()



"""Utilities to install packages for a cloned gist"""
from __future__ import print_function
import os
import re
import subprocess
from time import sleep
from sys import stderr
from jovian.utils.anaconda import get_conda_bin, CONDA_NOT_FOUND
from jovian.utils.misc import get_platform
from jovian.utils.constants import LINUX, WINDOWS, MACOS, ISSUES_MSG
from jovian.utils.logger import log


YML_PKG_LINE = r'^\s*-\s*([a-zA-Z0-9._-]+)(\s*==?\s*.*)?\s*$'

YML_NAME_LINE = r'^\s*name\s*:\s*([a-zA-Z0-9._-]+)\s*$'


blacklist = [
    'conda',
    'conda-env',
    'attrs'
]


def identify_env_file(env_fname):
    """Find the right conda environment file through trial and errors"""
    if env_fname is None:
        # Look for platform specific environment files (prefer current)
        platforms = [get_platform()] + ["", LINUX, WINDOWS, MACOS]
        for platform in platforms:
            if platform == "":
                expected_fname = 'environment.yml'
            else:
                expected_fname = 'environment-' + platform + '.yml'
            if os.path.exists(expected_fname):
                env_fname = expected_fname
                break

    if env_fname is None:
        # Check for standard environment.yml file
        if os.path.exists('environment.yml'):
            env_fname = 'environment.yml'

    return env_fname


def extract_pkg(line):
    """Extract the name of a package from a YML package line"""
    res = re.findall(YML_PKG_LINE, line)
    return res[0][0] if len(res) > 0 else None


def extract_env_name(env_fname):
    """Extract the name of the environment from the env file"""
    name = None
    with open(env_fname) as f:
        lines = f.read().split('\n')
        for line in lines:
            res = re.findall(YML_NAME_LINE, line)
            if len(res) > 0:
                name = res[0]
                break
    return name


def write_env_name(env_name, env_fname):
    """Overwrite the environement name into the file"""
    with open(env_fname) as f:
        # Read the environment file
        lines = f.read().split('\n')
        for i, line in enumerate(lines):
            # Check if it matches the name line pattern
            res = re.findall(YML_NAME_LINE, line)
            if len(res) > 0:
                # Extract the name
                name = res[0]
                # Find its poisition in the line
                idx = line.find(name)
                if idx != -1:
                    # Replace the name
                    lines[i] = line[:idx] + env_name + line[idx+len(name):]
    # Save env file with new name
    out_str = '\n'.join(lines)
    with open(env_fname, 'w') as f:
        f.write(out_str)


ENV_NAME_MSG = "Please provide a name for the conda environment"


def request_env_name(env_name, env_fname):
    """Request the user to provide a name for the environment"""
    if env_name is None:
        env_name = extract_env_name(env_fname)
        # Make sure we're not overwriting the base environment
        if env_name == 'base':
            env_name = None
        # Construct the help message with default value
        if env_name:
            msg = ENV_NAME_MSG + " [" + env_name + "]: "
        else:
            msg = ENV_NAME_MSG + ":"
        # Prompt the user for input
        try:
            user_input = raw_input(msg)
        except NameError:
            user_input = input(msg)
        print('')
        # Sanitize the input
        user_input = user_input.strip()
        # Set the final env name
        if user_input:
            env_name = user_input
        if env_name:
            # Write the chosen name to file
            write_env_name(env_name, env_fname)
    return env_name


def check_notfound(error_str):
    """Check if the error output contains ResolvePackageNotFound"""
    error_lines = error_str.split('\n')
    notfound = False
    pkgs = []
    for line in error_lines:
        if 'ResolvePackageNotFound:' in line:
            notfound = True
        if notfound:
            pkg = extract_pkg(line)
            if pkg:
                pkgs.append(pkg)
    return notfound, pkgs


def check_unsatisfiable(error_str):
    """Check if the error output contains ResolvePackageNotFound"""
    error_lines = error_str.split('\n')
    unsatisfiable = False
    pkgs = []
    for line in error_lines:
        if 'UnsatisfiableError:' in line:
            unsatisfiable = True
        if unsatisfiable:
            pkg = extract_pkg(line)
            if pkg:
                pkgs.append(pkg)
    return unsatisfiable, pkgs


def sanitize_envfile(env_fname, pkgs):
    """Remove the given packages from the environment file"""
    with open(env_fname) as f:
        # Read the lines of the environment file
        lines = f.read().split('\n')
        for i, line in enumerate(lines):
            # Extract package name (if present)
            pkg = extract_pkg(line)
            if pkg and (pkg in pkgs or pkg in blacklist):
                # Comment it out
                lines[i] = "# " + line
    # Save env file with pkgs commented out
    out_str = '\n'.join(lines)
    with open(env_fname, 'w') as f:
        f.write(out_str)


MISSING_MSG = ("WARNING: Some packages listed in the environment definition file could" +
               " not be installed, possibly because the environment was recorded on a different" +
               " operating system. As a result, you have to install some packages manually using " +
               "'conda install <package_name>' if you face errors while executing the code.\n")


def install(env_fname=None, env_name=None):
    """Install packages for a cloned gist"""
    # Check for conda and get the binary path
    conda_bin = get_conda_bin()

    # Identify the right environment file, and exit if absent
    env_fname = identify_env_file(env_fname)
    if env_fname is None:
        log('Failed to detect a conda environment YML file. Skipping..', error=True)
        return
    else:
        log('Detected conda environment file: ' + env_fname + "\n")

    # Get the environment name from user input
    env_name = request_env_name(env_name, env_fname)
    if env_name is None:
        log('Environment name not provided/detected. Skipping..')
        return

    # Construct the command
    command = conda_bin + ' env update --file "' + \
        env_fname + '" --name "' + env_name + '"'

    # Run the command
    log('Executing:\n' + command + "\n")
    install_task = subprocess.Popen(
        command, shell=True, stderr=subprocess.PIPE)

    # Extract the error (if any)
    _, error_str = install_task.communicate()
    error_str = error_str.decode('utf8', errors='ignore')

    if error_str:
        print(error_str, file=stderr)

        # Check for ResolvePackageNotFound error
        notfound, pkgs = check_notfound(error_str)

        # Sanitize the environment file if required
        if notfound:
            log('Installation failed!', error=True)
            log('Ignoring unresolved depedencies and trying again...\n')
            sleep(1)
            sanitize_envfile(env_fname, pkgs)

            # Try to install again
            log('Executing:\n' + command + "\n")
            install_task2 = subprocess.Popen(
                command, shell=True, stderr=subprocess.PIPE)

            # Extract the error (if any)
            _, error_str2 = install_task2.communicate()
            error_str2 = error_str2.decode('utf8', errors='ignore')
            if error_str2:
                print(error_str2, file=stderr)

                # Check for UnsatisfiableError
                unsatisfiable, pkgs2 = check_unsatisfiable(error_str2)

                # Sanitize the environement file further
                if unsatisfiable:
                    log('Installation failed!', error=True)
                    log('Ignoring unsatisfiable depedencies and trying again...\n')
                    sleep(1)
                    sanitize_envfile(env_fname, pkgs2)

                    # Try to install again
                    log('Executing:\n' + command + "\n")
                    install_task3 = subprocess.Popen(
                        command, shell=True, stderr=subprocess.PIPE)

                    # Extract the error (if any)
                    _, error_str3 = install_task3.communicate()
                    error_str3 = error_str3.decode('utf8', errors='ignore')
                    if error_str3:
                        print(error_str3, file=stderr)
                    else:
                        # Display a warning that some packages may be missing
                        log(MISSING_MSG)

    # Print beta warning and github link
    log(ISSUES_MSG)


def activate(env_fname=None):
    """Read the conda environment file and activate the environment"""
    # Check for conda and get the binary path
    try:
        conda_bin = get_conda_bin()
    except:
        log(CONDA_NOT_FOUND, error=True)
        return False

    # Identify the right environment file, and exit if absent
    env_fname = identify_env_file(env_fname)
    if env_fname is None:
        log('Failed to detect a conda environment YML file. Skipping..', error=True)
        return False
    else:
        log('Detected conda environment file: ' + env_fname + "\n")

    # Get the environment name from user input
    env_name = extract_env_name(env_fname)
    if env_name is None:
        log('Environment name not provided/detected. Skipping..')
        return False

    # Activate the environment
    command = conda_bin + " activate " + env_name
    log('Executing:\n' + command + "\n")
    task = subprocess.Popen(
        command, shell=True, stderr=subprocess.PIPE)

    # Extract the error (if any)
    _, error_str = task.communicate()
    error_str = error_str.decode('utf8', errors='ignore')
    print(error_str)

    # TODO: Try again with `source` for older versions of conda
    # Need to check it across platforms

    # Print beta warning and github link
    log(ISSUES_MSG)
